#include "CyanBus.hpp"

// C++ system level
#include <cstring>  // memset
#include <cstdlib>  // TODO: is this for memcpy?
#include <cinttypes>
#include <functional>

// ESP32 specific
#include "esp_log.h"
#include "driver/gpio.h"
#include "driver/uart.h"
#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"

// project specific
#include "LambdaTask.hpp"

// qthing stuff
// #include ""



static const char* TAG = "CyanBus";

#include "esp_err.h"
namespace SiliconTorch {

//   namespace Impl {
// 
//     typedef struct {
//       bool crcOK;
//       uint16_t length;
//       uint8_t payload[CyanBus::MTU];
//     } CyanData;
// 
//   }

  namespace CyanBus {

    // CyanBus envelope header
    const uint8_t* const HEADER = (const uint8_t*)"fxCyan";


    CyanBus::CyanBus(uint8_t tx, uint8_t rx, uint8_t de, uint8_t re, uint32_t baudRate, uint8_t uartChannel) : tx(tx), rx(rx), de(de), re(re), uartChannel(uartChannel) {


      uint64_t bitMask = 0;

      bitMask |= 1 << tx;
      bitMask |= 1 << de;
      bitMask |= 1 << re;

      gpio_config_t gpio = {
        .pin_bit_mask = bitMask,
        .mode         = GPIO_MODE_OUTPUT,
        .pull_up_en   = GPIO_PULLUP_DISABLE,
        .pull_down_en = GPIO_PULLDOWN_DISABLE,
        .intr_type    = GPIO_INTR_DISABLE,        
      };

      // configure outputs
      gpio_config(&gpio);
      txrxEN(false);

      gpio.pin_bit_mask = 1 << rx;
      gpio.mode = GPIO_MODE_INPUT;

      // configure inputs
      gpio_config(&gpio);


      uart_config_t uartConfig = {
        .baud_rate           = (int)baudRate,
        .data_bits           = UART_DATA_8_BITS,
        .parity              = UART_PARITY_DISABLE,
        .stop_bits           = UART_STOP_BITS_1,
        .flow_ctrl           = UART_HW_FLOWCTRL_DISABLE,
        .rx_flow_ctrl_thresh = 0,
        .use_ref_tick        = false,
        // .source_clk = UART_SCLK_APB,
      };


      uart_port_t _ch = (uart_port_t)uartChannel;

      ESP_ERROR_CHECK(uart_param_config(_ch, &uartConfig));
      ESP_ERROR_CHECK(uart_set_pin(_ch, tx, rx, UART_PIN_NO_CHANGE, UART_PIN_NO_CHANGE));
      //uart_intr_config(_ch);

      ESP_ERROR_CHECK(uart_driver_install(_ch, MTU * 2, 0, 20, &uartEvQ, 0));  // txBuf = 0; queue size = 32;  TODO: evaluate!!!

      ESP_ERROR_CHECK(uart_set_mode(_ch, UART_MODE_UART));

      // WARNING: The following call breaks EVERYTHING!
      uart_set_rx_timeout(_ch, 5);


      // TOOD: receive only ATM!
      txEN(false);
      rxEN(true);


      readerTaskHandle = new Util::LambdaTask([&](){ readerTask(); });

      // packetQ = xQueueCreate(3, sizeof(Impl::CyanData));
      packetQ = xQueueCreate(3, 4096);
      packetTaskHandle = new Util::LambdaTask([&](){ packetTask(); });
    }

    void CyanBus::readerTask() {

      uart_port_t _ch = (uart_port_t)uartChannel;

      uint32_t bufLEN = MTU;
      uint8_t* buffer = new uint8_t[bufLEN];  // our packet buffer
      uint8_t* bufPTR = buffer;               // points inside our buffer; used for envelope parsing…
      // uint16_t payloadLEN = 0;

      std::memset(buffer, 0x00, bufLEN);


      /* enum FSM { SearchHeader, ReadLength, ReadPayload, CheckCRC };
      FSM fsm = SearchHeader;

      uint8_t headerLength = std::strlen(HEADER);
      uint32_t headerByteNum = 0; */

      uart_event_t event;

      auto Q = uartEvQ;
      auto Q2 = packetQ;

      while (true) {

        if (xQueueReceive(Q, (void*)&event, (TickType_t)portMAX_DELAY)) {  // 0 = return immediately!

          auto msgs  = uxQueueMessagesWaiting(Q);
          auto space = uxQueueSpacesAvailable(Q);

          size_t available;
          uart_get_buffered_data_len(_ch, &available);

          ESP_LOGI(TAG, "Got a new UART event! 🤩   QByte[ %d ]  QMsg#[ %d ]  event.size[ %d ]  uart.available[ %d ]", space, msgs, event.size, available);

          switch (event.type) {

            // received plain data
            case UART_DATA: {

              uint32_t bytesRead = bufPTR - buffer;

              if (bytesRead + event.size > bufLEN) {
                ESP_LOGE(TAG, "Buffer#[ %d ] Overflow: Got %d bytes without a break", bufLEN, bytesRead + event.size);

                if (event.size < bufLEN) {  // read into buffer and just start over

                  uart_read_bytes(_ch, buffer, event.size, portMAX_DELAY);
                  bufPTR = buffer;

                } else {  // kill it with 🔥


                  ESP_LOGE(TAG, "// kill it with 🔥");
                  uart_flush_input(_ch);
                  bufPTR = buffer;

                }
              } else {
                
                /*int32_t _bytes = 0;
                do {
                  _bytes = uart_read_bytes(_ch, bufPTR, 16, 1);
                  ESP_LOGW(TAG, "_bytes#[ %d ]", _bytes);

                  if (_bytes < 0) {
                    ESP_LOGE(TAG, "!! _bytes < 0 !!");
                    break;
                  }
                  bufPTR += _bytes;
                } while (_bytes > 0);*/

                bufPTR += uart_read_bytes(_ch, bufPTR, event.size, portMAX_DELAY);
              }

              break;
            }

            // got an RX break signal
            case UART_BREAK: {
              ESP_LOGI(TAG, "Received BREAK 🛑 after reading %d bytes", bufPTR - buffer);


              char strbuf[65];
              strbuf[64] = 0x00;

              std::memcpy(strbuf, buffer, 64);

              ESP_LOGW(TAG, "PACKET DEBUG: Starts with [ %s ]", strbuf);


              // xQueueSendToBack(Q2, buffer, portMAX_DELAY);
              bufPTR = buffer;
              std::memset(buffer, 0x00, bufLEN);  // TODO: needed…?
              uart_flush_input(_ch);
              xQueueReset(Q);

              break;
            }

            default: {

              char* eventName;

              switch (event.type) {
                case UART_DATA:        { eventName = (char*)"UART_DATA";        break; }
                case UART_BREAK:       { eventName = (char*)"UART_BREAK";       break; }
                case UART_BUFFER_FULL: { eventName = (char*)"UART_BUFFER_FULL"; break; }
                case UART_FIFO_OVF:    { eventName = (char*)"UART_FIFO_OVF";    break; }
                case UART_FRAME_ERR:   { eventName = (char*)"UART_FRAME_ERR";   break; }
                case UART_PARITY_ERR:  { eventName = (char*)"UART_PARITY_ERR";  break; }
                case UART_DATA_BREAK:  { eventName = (char*)"UART_DATA_BREAK";  break; }
                case UART_PATTERN_DET: { eventName = (char*)"UART_PATTERN_DET"; break; }
                default:               { eventName = (char*)"_UNKNOWN_";        break; }
              }

              ESP_LOGE(TAG, "Unexpected event[ %s ]!", eventName);

              uart_flush_input(_ch);
              bufPTR = buffer;
              std::memset(buffer, 0x00, bufLEN);
            }
          }


        } else {
          // we should never get here. let's see if this holds ;)
          ESP_LOGE(TAG, "Receiving from queue failed…?!");
        }
      }

      delete buffer;
    }

    void CyanBus::packetTask() {
      ESP_LOGW(TAG, "TODO: implement callback task!");

      uint8_t* buffer = new uint8_t[MTU];

      auto Q = packetQ;
      while (true) {

        if (xQueueReceive(Q, buffer, (TickType_t)portMAX_DELAY)) {

          // TODO: handle packet!

          char strbuf[17];
          strbuf[16] = 0x00;

          std::memcpy(strbuf, buffer, 16);

          ESP_LOGI(TAG, "PACKET DEBUG: Starts with [ %s ]", strbuf);

          // TODO: really handle packet!
        }
      }
    }

    void CyanBus::setBaudRate(uint32_t baudRate) {
      uart_set_baudrate((uart_port_t)uartChannel, baudRate);
    }

    uint32_t CyanBus::getBaudRate() {
      uint32_t baudRate = 0;
      uart_get_baudrate((uart_port_t)uartChannel, &baudRate);
      return baudRate;
    }


    void CyanBus::txEN(bool state) {
      gpio_set_level((gpio_num_t)de, state);
    }

    void CyanBus::rxEN(bool state) {
      gpio_set_level((gpio_num_t)re, state ^ 1);
    }
    
    void CyanBus::txrxEN(bool state) {
      txEN(state);
      rxEN(state);
    }

  }
}



